/*
 * xreplot - Function plotting utility
 * Copyright(c) David Drysdale 1995-1996
 * Freely distributable provided this copyright notice remains intact
 *
 * Send bug reports to dmd@lurklurk.org
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "getopt.h"

#define XREPLOT

#ifdef XREPLOT
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Core.h>

#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/AsciiSrc.h>

#include "xreplot.bmp"
#include "xreploto.bmp"

#include "xutil.h"
#include "xreplot.h"

#include "xfunc_ps.h"
#include "xfunc_x.h"

#endif


/****************************************************************************/
/* Global variables                                                         */
/****************************************************************************/
FILE   *infile=NULL;
FILE   *outfile=NULL;
float  **data=NULL;
float  **dataimag=NULL;
float  *timelist=NULL;
float  timestep=1.0;
float  firsttimestep=-2.0;
float  pageoffset=0.0;
float  lasttickprinttime;
float  lasttickprint;
float  lastprint=-MINTIME;
float  nowtime=0.0;
int    norows=0;
int    nopix;
float  lastoffset;
char   options[256];
float  voffset=VOFFSET;  /* current vertical offset */
float  voffmax=VOFFMAX;  /* maximum voffset before reset */
float  vstep=VSTEP;    /* amount to increase voffset per step */
float  pagewidth=PAGEWIDTH;/* width of the page */
float  deltax;   /* amount to move right per x gridpoint */
float  uscaling=SCALING;  /* norm scaling factor as set by user */
float  scaling=SCALING;  /* norm scaling factor */
float  originx=ORIGINX;  /* initial origin translate */
float  originy=ORIGINY;
float  tick=TICK;  /* tick widths */
float  fontsize=FONTSIZE; /* font point size */
float  maxnorm=MAXNORM; /* norm scale */
float  printtime=PRINTTIME; /* when to print time ticks */
float  mintime=MINTIME; /* Minimum time separation between printed lines */
int    nopages=-1;
int    sample=1;
int    grid=1;
float  doubled=1.0;
int    idoubled=1;
int    lr=1;
int    complexin=FALSE;
int    grayscale=0;
long   whichpage=1;
int    reverse=1;
#ifdef XREPLOT
int    hardcopy=FALSE;
Display *display = NULL;
#else
int    hardcopy=TRUE;
#endif
char   infilename[1024];
char   outfilename[1024] = "";
int    ii;


/****************************************************************************/
/* Now define pointers to the output functions                              */
/****************************************************************************/
void (*startlinefunc)(float, float, float);
void (*linepointfunc)(float, float, float, float);
void (*endlinefunc)();
void (*startlinefunC)(float, float, float, float, float, float);
void (*linepointfunC)(float, float, float, float, float);
void (*endlinefunC)();
void (*printtickfunc)(float, float, float, float, float, int);
void (*prologfunc)(char *, int, float, float, float, float *, float, float,
                   float, float, float, float, float, int, int, int);
void (*eopagefunc)(float, float, float, float, float, float,
                   float, int, int, int);
void (*epilogfunc)(int, float, int, float, float);


/****************************************************************************/
/* printusage : Print out usage message                                     */
/****************************************************************************/
void printusage(int err)
{
  fprintf(stderr, "Usage: xreplot [options] filename\n");
  fprintf(stderr, "Creates a visualization of the data in filename.\n");
  fprintf(stderr, "Real data is assumed to be a list of numbers on each line, with\n");
  fprintf(stderr, "the timestep value in quotes at the beginning of the line.\n");
  fprintf(stderr, "Complex data (-c option) is assumed to be one complex number per\n");
  fprintf(stderr, "line, in form (real,imag), followed by a line with the timestep\n");
  fprintf(stderr, "value on it. Options are\n");
  fprintf(stderr, "   -a a    x coordinate for bottom left of plot (default %.2f)\n",
          originx);
  fprintf(stderr, "   -b b    y coordinate for bottom left of plot (default %.2f)\n",
          originy);
  fprintf(stderr, "   -y y    amount to vertically move between steps (default %.2f)\n",
          vstep);
  fprintf(stderr, "   -t t    maximum y coordinate for each page (default %.2f)\n",
          voffmax);
  fprintf(stderr, "   -r r    maximum x coordinate for each line (default %.2f)\n",
          pagewidth);
  fprintf(stderr, "   -n n    scaling factor for norm of function (default %.2f)\n",
          uscaling);
  fprintf(stderr, "   -f f    font size for labels (default %.2f)\n", fontsize);
  fprintf(stderr, "   -m m    size of norm scale at side (default %.2f)\n", maxnorm);
  fprintf(stderr, "   -p p    how often to print timestep labels (default %.2f)\n",
          printtime);
  fprintf(stderr, "   -s s    line sampling rate (default %d)\n", sample);
  fprintf(stderr, "   -x x    minimum time between plots (default %.2f)\n", mintime);
  fprintf(stderr, "   -k k    gridpoint sampling rate (default %d)\n", grid);
  fprintf(stderr, "   -d d    put two columns onto one page\n");
  fprintf(stderr, "   -g      produce gray scale output\n");
  fprintf(stderr, "   -i      produce inverse gray scale output\n");
  fprintf(stderr, "   -c      allow complex input (needs -w option)\n");
  fprintf(stderr, "   -e e    how many pages to output\n");
  fprintf(stderr, "   -q      reverse output\n");
  fprintf(stderr, "   -H      produce PostScript output\n");
  fprintf(stderr, "Units are in printers points (1/72 inch) for coordinates\n");
  exit(err);
}

/****************************************************************************/
/* find_width : Find the width of the data in the input file.  Note that    */
/* the file may be complex or real numbers                                  */
/****************************************************************************/
int find_width(FILE *infile, int complexin)
{
  int byte;
  int norows=0;

  rewind(infile);
  if (complexin) {
    while (isspace(byte = fgetc(infile))) ;
    while (byte=='(') {
      while ((byte != '\n') && (byte != EOF)) {
        byte = fgetc(infile);
      }
      /* slurp up leading whitespace on next line */
      while (isspace(byte = fgetc(infile))) ;
      norows++;
    }
  } else {
    /************************************************************************/
    /* find out the number of numbers on the first line.  assumed to be the */
    /* same for all subsequent lines                                        */
    /************************************************************************/
    while (isspace(byte = fgetc(infile))) ;
    if (byte != '\"') {
      fprintf(stderr, "Warning: expecting timestep tag on line 1\n");
    } else {
      byte = fgetc(infile);
      while ((byte != '\"') && (byte != '\n') && (byte != EOF)) {
        byte = fgetc(infile);
      }
      if (byte == '\n') {
        fprintf(stderr, "Error: line break in tag on line 1\n");
        exit(1);
      }
      else if (byte == EOF) {
        fprintf(stderr, "Error: file ended in tag on line 1\n");
        exit(1);
      }
      byte = fgetc(infile);
      while (isspace(byte = fgetc(infile))) ;
    }
    ungetc(byte, infile);
    while ( (byte != EOF) && (byte != '\n')) {
      norows++;
      while (!isspace(byte)) byte=fgetc(infile);
      while (isspace(byte) && byte != '\n') byte=fgetc(infile);
    }
  }
  return(norows);
}

/****************************************************************************/
/* get_data : Read in all the data into memory, allocating the necessary    */
/* arrays                                                                   */
/****************************************************************************/
int get_data(FILE *infile, int s, int k, int *c, int *n, float *lastoffset,
             float **timelist, float ***data, float ***dataimag)
{
  int nolines = 0;
  int alllines;
  int byte;
  char eof = 0;
  float temp1, temp2;
  float nowtime;
  int width;
  float timestep=1.0;
  int w;
  int ii;

  /**************************************************************************/
  /* First work out if the file has real or complex data in it              */
  /**************************************************************************/
  rewind(infile);
  while (isspace(byte=fgetc(infile)));
  *c = (byte=='(');
  rewind(infile);

  /**************************************************************************/
  /* Next work out the number of sets of data in the file                   */
  /**************************************************************************/
  rewind(infile);
  w = find_width(infile, *c);
  rewind(infile);
  while ( (byte=fgetc(infile))!=EOF )
    if (byte == '\n') nolines++;
  if (*c==1) {
    assert(nolines % (w+1) == 0);
    nolines = nolines / (w+1);
  }

  /* get the space */
  alllines = nolines;
  nolines = (nolines + s - 1) / s;

  /**************************************************************************/
  /* there are w points, thus (w-1) intervals, thus (w-1)/k chunks round    */
  /* this up to get the size of array needed, take the fractional part to   */
  /* get the ration of the last point                                       */
  /**************************************************************************/
  temp1 = ((float) (w-1)/(float) k);
  width = (int)temp1;
  if ((float) width == temp1) {
    (*lastoffset) = 1.0;
  } else {
    (*lastoffset) = temp1 - (float) width;
    width++;
  }
  width++;
  (*timelist) = vector(0, nolines-1);
  (*data) = matrix(0, nolines-1, 0, width-1);
  if (*c==1)
    (*dataimag) = matrix(0, nolines-1, 0, width-1);

  /**************************************************************************/
  /* Now actually read in the data                                          */
  /**************************************************************************/
  rewind(infile);
  eof = 0;
  *n = nolines;
  nolines = 0;
  if (*c==1) {
    /* complex input case */
    while ((!eof) && (nolines < alllines)) {
      if (nolines % s == 0) {
        for (ii=0; ii<w; ii++) {
          if (((ii % k) == 0) || (ii==(w-1))) {
            eof |= (fscanf(infile, "(%f, %f)\n",
                              &((*data)[nolines/s][(ii+k-1)/k]),
                              &((*dataimag)[nolines/s][(ii+k-1)/k]) ) != 2);
            assert(!eof);
          } else {
            eof |= (fscanf(infile, "(%f, %f)\n", &temp1, &temp2) != 2);
          }
        }
        eof |= (fscanf(infile, "%f\n", &((*timelist)[nolines/s])) != 1);
      } else {
        for (ii=0; ii<(w+1); ii++) {
          byte = '\0';
          while ((byte != '\n') && (byte != EOF))
            byte = fgetc(infile);
        }
        eof = (byte == EOF);
      }
      nolines++;
    }
  } else {
    /* real input case */
    nowtime = -timestep;
    while ((!eof) && (nolines < alllines)) {
      if (nolines % s == 0) {
        while (isspace(byte = fgetc(infile))) ;
        if (byte != '\"') {
          ungetc(byte, infile);
          nowtime = (*timelist)[nolines/s] = nowtime+timestep;
        } else {
          eof |= (fscanf(infile, "%f\"", &((*timelist)[nolines/s])) != 1);
          nowtime = (*timelist)[nolines/s];
        }
        for (ii=0; ii<w; ii++) {
          if (((ii % k) == 0) || (ii==(w-1))) {
            eof |= (fscanf(infile, "%f",
                           &((*data)[nolines/s][(ii+k-1)/k])) != 1);
          } else {
            eof |= (fscanf(infile, "%f", &temp2) != 1);
          }
        }
      } else {
        byte = '\0';
        while ((byte != '\n') && (byte != EOF)) {
          byte = fgetc(infile);
        }
        eof = (byte == EOF);
      }
      nolines++;
    }
  }
  return(width);
}

/****************************************************************************/
/* Function to set global variables from options.  Only called once         */
/****************************************************************************/
void set_options(int *argc, char **argv[])
{
  int opt;
  while ((opt = getopt(*argc,*argv,"a:b:cd:e:f:ghik:m:n:p:qr:s:t:x:y:H")) != EOF)
    switch (opt)
    {
      case 'a': originx = atof(optarg); break;
      case 'b': originy = atof(optarg); break;
      case 'c': complexin = TRUE; break;
      case 'd': idoubled = atoi(optarg);
                doubled = (float) idoubled;
                if (doubled < 1.0) doubled = 1.0;
                break;
      case 'e': nopages=atoi(optarg); break;
      case 'f': fontsize = atof(optarg); break;
      case 'g': grayscale = 1; break;
      case 'h': printusage(0); break;
      case 'i': grayscale = -1; break;
      case 'k': grid = atoi(optarg);
                if (grid <= 0) grid = 1;
                break;
      case 'm': maxnorm = atof(optarg); break;
      case 'n': uscaling = scaling = atof(optarg); break;
      case 'p': printtime = atof(optarg); break;
      case 'q': reverse=-1; break;
      case 'r': pagewidth = atof(optarg); break;
      case 's': sample = atoi(optarg);
                if (sample <= 0) sample = 1;
                break;
      case 't': voffmax = atof(optarg); break;
      case 'x': mintime = atof(optarg);
                lastprint = -mintime;
                break;
      case 'y': vstep = atof(optarg); break;
      case 'H': hardcopy = TRUE; break;
      default: printusage(2); break;
    }
  *argc -= optind;
  *argv += optind;

  if (grayscale) {
    scaling = uscaling / GRAYSCALE;
  }
  if (*argc>=1)
    strcpy(infilename, (*argv)[0]);
}

/****************************************************************************/
/* Point the plotting functions at the right versions                       */
/****************************************************************************/
void set_plotfuncs()
{
  if (hardcopy) {
    if (complexin) {
      if (grayscale) {
        startlinefunC = &ps_complex_gray_sl;
        linepointfunC = &ps_complex_gray_lp;
        endlinefunC = &ps_complex_gray_el;
      } else {
        startlinefunC = &ps_complex_func_sl;
        linepointfunC = &ps_complex_func_lp;
        endlinefunC = &ps_complex_func_el;
      }
    } else {
      if (grayscale) {
        startlinefunc = &ps_real_gray_sl;
        linepointfunc = &ps_real_gray_lp;
        endlinefunc = &ps_real_gray_el;
      } else {
        startlinefunc = &ps_real_func_sl;
        linepointfunc = &ps_real_func_lp;
        endlinefunc = &ps_real_func_el;
      }
    }
    prologfunc = &ps_prolog;
    eopagefunc = &ps_eopage;
    epilogfunc = &ps_epilog;
    printtickfunc = &ps_printtick;
#ifdef XREPLOT
  } else {
    if (complexin) {
      prologfunc = &x_complex_prolog;
      if (grayscale) {
        startlinefunC = &x_complex_gray_sl;
        linepointfunC = &x_complex_gray_lp;
        endlinefunC = &x_complex_gray_el;
      } else {
        startlinefunC = &x_complex_func_sl;
        linepointfunC = &x_complex_func_lp;
        endlinefunC = &x_complex_func_el;
      }
    } else {
      prologfunc = &x_real_prolog;
      if (grayscale) {
        startlinefunc = &x_real_gray_sl;
        linepointfunc = &x_real_gray_lp;
        endlinefunc = &x_real_gray_el;
      } else {
        startlinefunc = &x_real_func_sl;
        linepointfunc = &x_real_func_lp;
        endlinefunc = &x_real_func_el;
      }
    }
    eopagefunc = &x_eopage;
    epilogfunc = &x_epilog;
    printtickfunc = &x_printtick;
#endif
  }
}


/****************************************************************************/
/* Generic routine for doing the displaying.  This is independent of        */
/* whether the output format is to the screen or postscript                 */
/****************************************************************************/
void OnExpose()
{
  float  temp;
  char   byte;
  int    ii;
  int    ll;
  float  theta;

  /**************************************************************************/
  /* first prolog                                                           */
  /**************************************************************************/
  (*prologfunc)(options, norows, voffmax, vstep, pagewidth, &deltax,
                scaling, originx, originy, tick, fontsize, maxnorm,
                doubled, complexin, grayscale, reverse);
  if (reverse==-1) {
    vstep = -fabs(vstep);
  } else {
    vstep = fabs(vstep);
  }

  /**************************************************************************/
  /* reset all variables                                                    */
  /**************************************************************************/
  lr = 1;
  whichpage = 1;
  nowtime = 0.0;
  pageoffset = 0.0;
  voffset = 0.0;
  firsttimestep = -2.0;
  lastprint = -mintime;

  /**************************************************************************/
  /*                         ** main loop **                                */
  /**************************************************************************/
  for (ll=0; ll<nopix; ll++) {
    /************************************************************************/
    /* Work out timestep information                                        */
    /************************************************************************/
    timestep = timelist[ll] - nowtime;
    nowtime = timelist[ll];
    if (firsttimestep == -2.0) {
      /* the first data -- set up things */
      lasttickprinttime=(float)((int)(nowtime/printtime)) * printtime;
      lasttickprint = lasttickprinttime-nowtime;
      timestep = 0.0;
      firsttimestep = -1.0;
     } else if (firsttimestep == -1.0) {
      firsttimestep = timestep;
     }

    /************************************************************************/
    /* Print out tickmarks                                                  */
    /************************************************************************/
    voffset +=  (vstep * (timestep/firsttimestep));
    while (nowtime >= (lasttickprinttime+printtime)) {
      lasttickprinttime+=printtime;
      lasttickprint+=printtime;
      temp=vstep*lasttickprint/(firsttimestep);
      temp-=pageoffset;
      if (fontsize != 0.0) {
        (*printtickfunc)(temp, lasttickprinttime, fontsize, tick,
                         doubled, grayscale);
      }
    }
    /************************************************************************/
    /* plot the line of data                                                */
    /************************************************************************/
    if (nowtime >= (lastprint+mintime)) {

      /**********************************************************************/
      /* Complex number version                                             */
      /**********************************************************************/
      if (complexin) {
        lastprint = nowtime;
        theta = (atan2(dataimag[ll][1], data[ll][1]) + PI) / (2.0 * PI);
        (*startlinefunC)(voffset,
                  ((data[ll][0]*data[ll][0]+
                    dataimag[ll][0]*dataimag[ll][0])*scaling),
                  theta,
                  (1*deltax)/doubled,
                  ((data[ll][1]*data[ll][1]+
                    dataimag[ll][1]*dataimag[ll][1])*scaling), vstep);
        for (ii=2; ii<(norows-1); ii++) {
          theta = (atan2(dataimag[ll][ii], data[ll][ii]) + PI) / (2.0 * PI);
          (*linepointfunC)(voffset, theta, (ii*deltax)/doubled,
                 (data[ll][ii]*data[ll][ii]+
                  dataimag[ll][ii]*dataimag[ll][ii])*scaling, vstep);
        }
        /* Always include the last point on the line */
        ii = (norows-1);
        theta = (atan2(dataimag[ll][ii], data[ll][ii]) + PI) / (2.0 * PI);
        (*linepointfunC)(voffset, theta,
                  (((norows-2)*deltax)/doubled)+(lastoffset*deltax/doubled),
                  (data[ll][ii]*data[ll][ii]+
                   dataimag[ll][ii]*dataimag[ll][ii])*scaling, vstep);
        (*endlinefunC)();

      /**********************************************************************/
      /* Real number version                                                */
      /**********************************************************************/
      } else {
        lastprint = nowtime;
        (*startlinefunc)(voffset, data[ll][0]*scaling, vstep);
        for (ii=1; ii<(norows-1); ii++) {
          (*linepointfunc)(voffset, (ii*deltax)/doubled, data[ll][ii]*scaling, vstep);
        }
        (*linepointfunc)(voffset,
                     (((norows-2)*deltax)/doubled)+(lastoffset*deltax/doubled),
                     data[ll][norows-1]*scaling, vstep);
        (*endlinefunc)();
      } /* not complexin */
    } /* time to print again */

    /************************************************************************/
    /* End of page code                                                     */
    /************************************************************************/
    if (fabs(voffset) >= voffmax) {
      lr++;
      if ((float) lr > doubled)
        lr = 1;
      if (lr == 1) {
        whichpage++;
      }

      (*eopagefunc)(voffset, pagewidth, vstep, doubled, originx, originy,
                    fontsize, grayscale, whichpage, lr);

      /* was this the last page we wanted to process */
      if ((whichpage > nopages) && (nopages != -1))
        return;

      /* is this other than the first page for an X display ? */
      if ((whichpage > 1) && (!hardcopy))
        return;
      pageoffset +=voffset;
      voffset =  -(vstep * (timestep/firsttimestep));
      /* voffset = 0.0; */
    } /* if voffset is high enough up */
  } /* while its not the end of the file */

  /**************************************************************************/
  /* output epilog                                                          */
  /**************************************************************************/
  (*epilogfunc)(grayscale, voffset, whichpage, pagewidth, doubled);
}

/****************************************************************************/
/* Widget callback functions                                                */
/****************************************************************************/
#ifdef XREPLOT
typedef struct tagWidgetValue {
  Widget w;
  int    type;
#define INT_TYPE 1
#define FLOAT_TYPE 2
#define STR_TYPE 3
  int    notrans;
  int *  iptr;
  float *fptr;
  char  *sptr;
} WidgetValue;
#define NUMWIDGETS 16
WidgetValue warray[NUMWIDGETS] = {0};
WidgetValue warrayl[NUMWIDGETS] = {0};
int warr_curr=0;
int warrl_curr=0;
  XtAppContext app_context;
  Widget topLevel, topLevelopt, viewwin;
  Pixmap icon_pixmap, icon_pixmap_o;
  Widget optbox, redraw, quit, pvshell; 
  Widget boxin, boxfnam, labifnm, infnam, comfnam;
  Widget boxbl, labbl, boxblx, labblx, txtblx, boxbly, labbly, txtbly;
  Widget boxtr, labtr, boxtrx, labtrx, txttrx, boxtry, labtry, txttry;
  Widget boxstep, labstep, txtstep, boxnorm, labnorm, txtnorm;
  Widget boxnscl, labnscl, txtnscl, boxfsiz, labfsiz, txtfsiz;
  Widget boxtick, labtick, txttick, boxtint, labtint, txttint;
  Widget boxmtin, labmtin, txtmtin, boxtsam, labtsam, txttsam;
  Widget boxxsam, labxsam, txtxsam, boxncol, labncol, txtncol;
  Widget boxmaxp, labmaxp, txtmaxp;
  Widget boxoutr, togfunc, toggray, togrgry, togtimr;
  Widget outfnam, printnow, boxofnm, labofnm;
  static int val [] = {0, 1, 2};
  static XtActionsRec actions[] = { {"DrawwinExpose", DrawwinExpose}, 
                                    {"UpdateValue",   UpdateValue},
                                    {"PrevField",     PrevField},
                                    {"NextField",     NextField},
                                    {"ThisField",     ThisField},
                                    {"ReturnActionP", ReturnActionP},
                                    {"ReturnActionS", ReturnActionS},
                                    {"ReturnAction",  ReturnAction}};
  String transTable = "<Expose>:        DrawwinExpose()";
  static char radiotranstab[] = "<Btn1Down>,<Btn1Up>:set() notify()";
  XtTranslations radiotrans;
  static char updatetranstab[] =
"~Shift<Key>Return:	UpdateValue() ReturnAction()\n\
Shift<Key>Return:	UpdateValue() ReturnAction()\n\
Shift<Key>Tab:		PrevField()\n\
<Key>Tab:		NextField()\n\
<Btn1Down>, <Btn1Up>:	ThisField()\n\
<FocusIn>:		display-caret(on)\n\
<FocusOut>:		display-caret(off)\n";
  static char updatetranstabl[] =
"~Shift<Key>Return:	ReturnActionS()\n\
Shift<Key>Return:	ReturnActionS()\n\
Shift<Key>Tab:		PrevField()\n\
<Key>Tab:		NextField()\n\
<Btn1Down>, <Btn1Up>:	ThisField()\n\
<FocusIn>:		display-caret(on)\n\
<FocusOut>:		display-caret(off)\n";
  XtTranslations updatetranslations;
#endif

void DrawwinExpose(Widget w, XEvent *event,
                   String *params, Cardinal *num_params)
{
  OnExpose();
}

void setupstring(int ii)
{
  char tempstr[256];
  if (warray[ii].type == INT_TYPE) {
    sprintf(tempstr, "%d", *(warray[ii].iptr));
  } else if (warray[ii].type == FLOAT_TYPE) {
    sprintf(tempstr, "%f", *(warray[ii].fptr));
  } else {
    return;
  }
  doubled = (float)idoubled;
  XtVaSetValues(warray[ii].w, XtNstring, tempstr, NULL);
}
void setupstringl(int ii)
{
  char tempstr[256];
  if (warrayl[ii].type == INT_TYPE) {
    sprintf(tempstr, "%d", *(warrayl[ii].iptr));
  } else if (warrayl[ii].type == FLOAT_TYPE) {
    sprintf(tempstr, "%f", *(warrayl[ii].fptr));
  } else if (warrayl[ii].type == STR_TYPE) {
    sprintf(tempstr, "%s", warrayl[ii].sptr);
  } else {
    return;
  }
  XtVaSetValues(warrayl[ii].w, XtNstring, tempstr, NULL);
}

void Redraw(Widget w, XtPointer client_data, XtPointer call_data)
{
  int ii;
  String string;

  for (ii=0; ii<warr_curr; ii++) {
    XtVaGetValues(warray[ii].w, XtNstring, &string, NULL);
    if (warray[ii].type == INT_TYPE) {
      *(warray[ii].iptr) = atoi(string);
    } else if (warray[ii].type == FLOAT_TYPE) {
      *(warray[ii].fptr) = atof(string);
    } else {
      continue;
    }
    setupstring(ii);
  }
  if (grayscale) {
    scaling = uscaling / GRAYSCALE;
  }
  XtVaSetValues(drawwin, XtNwidth, (int)(originx+pagewidth+18+4*fontsize),
                         XtNheight, (int)(originy + voffmax + 10*scaling),
                         NULL);
  XClearWindow(display, XtWindow(drawwin));
  OnExpose();
}

void ReturnAction(Widget w, XEvent *event,
                  String *params, Cardinal *num_params)
{
  Redraw(w, NULL, NULL);
}



void NextField(Widget w, XEvent *event,
               String *params, Cardinal *num_params)
{
  int ii;
  /* scan through the list of widgets to see which one this is */
  for (ii=0; ii<warr_curr; ii++) {
    if (warray[ii].w == w) {
      if (++ii!=warr_curr) {
        XSetInputFocus(display, XtWindow(warray[ii].w), RevertToParent, CurrentTime);
      } else {
        XSetInputFocus(display, XtWindow(warrayl[0].w), RevertToParent, CurrentTime);
      }
      break;
    }
  }
  if (ii==warr_curr) {
    for (ii=0; ii<warrl_curr; ii++) {
      if (warrayl[ii].w == w) {
        if (++ii!=warrl_curr) {
          XSetInputFocus(display, XtWindow(warrayl[ii].w), RevertToParent, CurrentTime);
        } else {
          XSetInputFocus(display, XtWindow(warray[0].w), RevertToParent, CurrentTime);
        }
        break;
      }
    }
  }
}

void PrevField(Widget w, XEvent *event,
               String *params, Cardinal *num_params)
{
  int ii;
  /* scan through the list of widgets to see which one this is */
  for (ii=0; ii<warr_curr; ii++) {
    if (warray[ii].w == w) {
      if (--ii!=-1) {
        XSetInputFocus(display, XtWindow(warray[ii].w), RevertToParent, CurrentTime);
      } else {
        XSetInputFocus(display, XtWindow(warrayl[warrl_curr-1].w), RevertToParent, CurrentTime);
      }
      break;
    }
  }
  if (ii==warr_curr) {
    for (ii=0; ii<warrl_curr; ii++) {
      if (warrayl[ii].w == w) {
        if (--ii!=-1) {
          XSetInputFocus(display, XtWindow(warrayl[ii].w), RevertToParent, CurrentTime);
        } else {
          XSetInputFocus(display, XtWindow(warray[warr_curr-1].w), RevertToParent, CurrentTime);
        }
        break;
      }
    }
  }
}

void ThisField(Widget w, XEvent *event,
               String *params, Cardinal *num_params)
{
  XSetInputFocus(display, XtWindow(w), RevertToParent, CurrentTime);
}

void Load(Widget w, XtPointer client_data, XtPointer call_data)
{
  String string;
  int ii;
  char tempstr[256];
  Position x, y;
  Dimension width, height;
  XEvent event;

  if (strlen(infilename) > 0) {
    if (infile!=NULL) fclose(infile);
    if (data!=NULL) free_matrix(data, 0, nopix-1, 0, norows-1);
    if (dataimag!=NULL) free_matrix(dataimag, 0, nopix-1, 0, norows-1);
    if (timelist!=NULL) free_vector(timelist, 0, nopix-1);
  }
  for (ii=0; ii<warrl_curr; ii++) {
    XtVaGetValues(warrayl[ii].w, XtNstring, &string, NULL);
    if (warrayl[ii].type == INT_TYPE) {
      *(warrayl[ii].iptr) = atoi(string);
    } else if (warrayl[ii].type == FLOAT_TYPE) {
      *(warrayl[ii].fptr) = atof(string);
    } else {
      continue;
    }
    setupstringl(ii);
  }
  XtVaGetValues(infnam, XtNstring, &string, NULL);
  strcpy(infilename, string);
  if ((infile = fopen(infilename, "r")) == NULL) {
    fprintf(stderr, "Error: can't open %s\n", infilename);
    exit(1);
  }
  norows = get_data(infile, sample, grid, &complexin,
                    &nopix, &lastoffset, &timelist, &data, &dataimag);
  set_plotfuncs();

  Redraw(0L, NULL,NULL);
}

void ReturnActionS(Widget w, XEvent *event,
                   String *params, Cardinal *num_params)
{
  Load(w, NULL, NULL);
}


void PrintNow(Widget w, XtPointer client_data, XtPointer call_data)
{
  String string;
  char tempstr[256];
  Position x, y;
  Dimension width, height;

  XtVaGetValues(outfnam, XtNstring, &string, NULL);
  strcpy(outfilename, string);
  if (outfile != NULL) fclose(outfile);
  if ((outfile = fopen(outfilename, "w")) == NULL) {
    fprintf(stderr, "Failed to open %s for writing\n", outfilename);
    return;
  }

  hardcopy = TRUE;
  set_plotfuncs();
  OnExpose();
  hardcopy = FALSE;
  set_plotfuncs();
}

void ReturnActionP(Widget w, XEvent *event,
                   String *params, Cardinal *num_params)
{
  PrintNow(w, NULL, NULL);
}


void UpdateValue(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  String string;
  int ii;

  XtVaGetValues(w, XtNstring, &string, NULL);
  for (ii=0; ii<warr_curr; ii++) {
    if (w == warray[ii].w) {
      if (warray[ii].type == INT_TYPE) {
        *(warray[ii].iptr) = atoi(string);
      } else if (warray[ii].type == FLOAT_TYPE) {
        *(warray[ii].fptr) = atof(string);
      } else {
        continue;
      }
      setupstring(ii);
      break;
    }
  }
}

void Quit(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (complexin)
    printf("-c ");
  if (grayscale==1)
    printf("-g ");
  else if (grayscale==-1)
    printf("-i ");
  printf( "-a %f -b %f -y %f -t %f -r %f -n %f -f %f -m %f -p %f "
          "-s %d -d %f -x %f -k %d\n", originx, originy, vstep, voffmax,
          pagewidth, uscaling, fontsize, maxnorm, printtime, sample, doubled,
          mintime, grid );
  printf("%s->%s\n", infilename, outfilename);
  exit(0);
}

void Outfunc(Widget w, XtPointer client_data, XtPointer call_data)
{
  long outtype=(long)client_data;
  switch (outtype) {
    case 0:
      scaling = uscaling;
      grayscale = 0;
      break;
    case 1:
      scaling = uscaling / GRAYSCALE;
      grayscale = 1;
      break;
    case 2:
      scaling = uscaling / GRAYSCALE;
      grayscale = -1;
      break;
  }
  set_plotfuncs();
}

void Reverse(Widget w, XtPointer client_data, XtPointer call_data)
{
  reverse = -reverse;
}

/****************************************************************************/
/* Main routine                                                             */
/****************************************************************************/
main(int argc, char * argv[])
{

  /**************************************************************************/
  /* check the command line arguments then record the options in a string   */
  /**************************************************************************/
  infilename[0] = (char) 0;
  set_options(&argc, &argv);
  sprintf(options, "-a %f -b %f -y %f -t %f -r %f -n %f -f %f -m %f -p %f "
          "-s %d -d %f -x %f -k %d\n", originx, originy, vstep, voffmax,
          pagewidth, uscaling, fontsize, maxnorm, printtime, sample, doubled,
          mintime, grid );

  /**************************************************************************/
  /* Open the file and get the data                                         */
  /**************************************************************************/
  if (strlen(infilename) > 0) {
    if ((infile = fopen(infilename, "r")) == NULL) {
      fprintf(stderr, "Error: can't open %s\n", infilename);
      exit(1);
    }
    norows = get_data(infile, sample, grid, &complexin,
                      &nopix, &lastoffset, &timelist, &data, &dataimag);
  }

  /**************************************************************************/
  /* set up the plotting functions                                          */
  /**************************************************************************/
  set_plotfuncs();

  /**************************************************************************/
  /* If we are just doing hardcopy, call the OnExpose function once and     */
  /* exit                                                                   */
  /**************************************************************************/
  if (hardcopy) {
    OnExpose();
    return(0);
  }

#ifdef XREPLOT
  /**************************************************************************/
  /* Create the top level widget for the drawing stuff                      */
  /**************************************************************************/
  topLevel = XtVaAppInitialize(&app_context,      /* Application context    */
                               "XReplot",         /* Application class      */
                               NULL, 0,           /* command line opt list  */
                               &argc, argv,       /* command line args      */
                               NULL,              /* missing app-deflts file*/
                               NULL);             /* terminate varargs list */

  /**************************************************************************/
  /* Connect an icon with this top level widget                             */
  /**************************************************************************/
  icon_pixmap = XCreateBitmapFromData(XtDisplay(topLevel),
                                      RootWindowOfScreen(XtScreen(topLevel)),
                                      (const char *)xreplot_bits,
                                      xreplot_width,
                                      xreplot_height);
  icon_pixmap_o = XCreateBitmapFromData(XtDisplay(topLevel),
                                        RootWindowOfScreen(XtScreen(topLevel)),
                                        (const char  *)xreploto_bits,
                                        xreploto_width,
                                        xreploto_height);
  XtVaSetValues(topLevel, XtNiconPixmap, icon_pixmap_o, NULL);

  /**************************************************************************/
  /* Create a viewport widget to do the drawing in                          */
  /**************************************************************************/
  pvshell = XtVaCreatePopupShell("xreplot",
                                 topLevelShellWidgetClass,
                                 topLevel,
                                 XtNallowShellResize, TRUE,
                                 NULL);
  XtVaSetValues(pvshell,
                XtNiconPixmap,
                icon_pixmap,
                NULL);
  viewwin = XtVaCreateManagedWidget(infilename,
                                    viewportWidgetClass,
                                    pvshell,
                                    XtNallowHoriz, TRUE,
                                    XtNallowVert, TRUE,
                                    NULL);

  voffmxx = originy+voffmax+10.0*scaling;
  drawwin = XtVaCreateManagedWidget("drawwin",
                                    coreWidgetClass,
                                    viewwin,
                                    XtNheight, (int) voffmxx,
                                    XtNwidth,
                                      (int) (pagewidth+originx+18+4*fontsize),
                                    XtNtranslations,
                                      XtParseTranslationTable(transTable),
                                    NULL);
  XtAppAddActions(app_context, actions, XtNumber(actions));
  XtPopup(pvshell, XtGrabNone);

  /**************************************************************************/
  /* Create a box widget with all the options in it                         */
  /**************************************************************************/


  optbox = XtVaCreateManagedWidget("optbox",
                                   boxWidgetClass,
                                   topLevel,
                                   NULL);

  boxfnam = XtVaCreateManagedWidget("boxfnam",
                                    formWidgetClass,
                                    optbox,
                                    NULL);

      labifnm = XtVaCreateManagedWidget("Filename:",
                                        labelWidgetClass,
                                        boxfnam,
                                        XtNborderWidth, 0,
                                        NULL);
      infnam = XtVaCreateManagedWidget("infnam",
                                       asciiTextWidgetClass,
                                       boxfnam,
                                       XtNeditType, XawtextEdit,
                                       XtNresizable, True,
                                       XtNresize, XawtextResizeBoth,
                                       XtNfromVert, labifnm,
                                       XtNright, XtChainRight,
                                       XtNleft, XtChainLeft,
                                       XtNdisplayCaret, False,
                                       NULL);
      warrayl[warrl_curr].w = infnam;
      warrayl[warrl_curr].type = STR_TYPE;
      warrayl[warrl_curr].sptr = infilename;
      setupstringl(warrl_curr++);

      printnow = XtVaCreateManagedWidget("Load",
                                         commandWidgetClass,
                                         boxfnam,
                                         XtNfromVert, infnam,
                                         NULL);
      XtAddCallback(printnow, XtNcallback, Load, NULL);
                                        
  boxofnm = XtVaCreateManagedWidget("boxofnm",
                                    formWidgetClass,
                                    optbox,
                                    NULL);

      labofnm = XtVaCreateManagedWidget("Output filename:",
                                        labelWidgetClass,
                                        boxofnm,
                                        XtNborderWidth, 0,
                                        NULL);
      outfnam = XtVaCreateManagedWidget("outfnam",
                                        asciiTextWidgetClass,
                                        boxofnm,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromVert, labofnm,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      /* this has a translation table all of its own */
      updatetranslations = XtParseTranslationTable(
"~Shift<Key>Return:	ReturnActionP()\n\
Shift<Key>Return:	ReturnActionP()\n\
Shift<Key>Tab:		PrevField()\n\
<Key>Tab:		NextField()\n\
<Btn1Down>, <Btn1Up>:	ThisField()\n\
<FocusIn>:		display-caret(on)\n\
<FocusOut>:		display-caret(off)\n");
      XtOverrideTranslations(outfnam, updatetranslations);
      warrayl[warrl_curr].w = outfnam;
      warrayl[warrl_curr].type = STR_TYPE;
      warrayl[warrl_curr].sptr = outfilename;
      warrayl[warrl_curr].notrans = 1;
      setupstringl(warrl_curr++);
      printnow = XtVaCreateManagedWidget("Print",
                                         commandWidgetClass,
                                         boxofnm,
                                         XtNfromVert, outfnam,
                                         NULL);
      XtAddCallback(printnow, XtNcallback, PrintNow, NULL);
                                        
                                  
  boxin = XtVaCreateManagedWidget("boxbl",
                                  formWidgetClass,
                                  optbox,
                                  NULL);
      boxtsam = XtVaCreateManagedWidget("boxtsam",
                                        formWidgetClass,
                                        boxin,
                                        XtNborderWidth, 0,
                                        NULL);
          labtsam = XtVaCreateManagedWidget("t sampling",
                                            labelWidgetClass,
                                            boxtsam,
                                            XtNborderWidth, 0,
                                            NULL);
          txttsam = XtVaCreateManagedWidget("txttsam",
                                            asciiTextWidgetClass,
                                            boxtsam,
                                            XtNeditType, XawtextEdit,
                                            XtNresizable, True,
                                            XtNresize, XawtextResizeBoth,
                                            XtNfromHoriz, labtsam,
                                            XtNright, XtChainRight,
                                            XtNleft, XtChainLeft,
                                            XtNdisplayCaret, False,
                                            NULL);
          warrayl[warrl_curr].w = txttsam;
          warrayl[warrl_curr].type = INT_TYPE;
          warrayl[warrl_curr].iptr = &sample;
          setupstringl(warrl_curr++);

      boxxsam = XtVaCreateManagedWidget("boxxsam",
                                        formWidgetClass,
                                        boxin,
                                        XtNborderWidth, 0,
                                        XtNfromVert, boxtsam,
                                        NULL);
          labxsam = XtVaCreateManagedWidget("x sampling",
                                            labelWidgetClass,
                                            boxxsam,
                                            XtNborderWidth, 0,
                                            NULL);
          txtxsam = XtVaCreateManagedWidget("txtxsam",
                                            asciiTextWidgetClass,
                                            boxxsam,
                                            XtNeditType, XawtextEdit,
                                            XtNresizable, True,
                                            XtNresize, XawtextResizeBoth,
                                            XtNfromHoriz, labxsam,
                                            XtNright, XtChainRight,
                                            XtNleft, XtChainLeft,
                                            XtNdisplayCaret, False,
                                            NULL);
          warrayl[warrl_curr].w = txtxsam;
          warrayl[warrl_curr].type = INT_TYPE;
          warrayl[warrl_curr].iptr = &grid;
          setupstringl(warrl_curr++);

  boxbl = XtVaCreateManagedWidget("boxbl",
                                  formWidgetClass,
                                  optbox,
                                  NULL);
      labbl = XtVaCreateManagedWidget("Bottom left of graph",
                                      labelWidgetClass,
                                      boxbl,
                                      XtNborderWidth, 0,
                                      NULL);
      boxblx = XtVaCreateManagedWidget("boxblx",
                                       formWidgetClass,
                                       boxbl,
                                       XtNfromVert, labbl,
                                       XtNborderWidth, 0,
                                       NULL);
          labblx = XtVaCreateManagedWidget("x=",
                                           labelWidgetClass,
                                           boxblx,
                                           XtNborderWidth, 0,
                                           NULL);
          txtblx = XtVaCreateManagedWidget("txtblx",
                                           asciiTextWidgetClass,
                                           boxblx,
                                           XtNeditType, XawtextEdit,
                                           XtNresizable, True,
                                           XtNresize, XawtextResizeBoth,
                                           XtNfromHoriz, labblx,
                                           XtNright, XtChainRight,
                                           XtNleft, XtChainLeft,
                                           XtNwidth, 80,
                                           XtNdisplayCaret, False,
                                           NULL);
          warray[warr_curr].w = txtblx;
          warray[warr_curr].type = FLOAT_TYPE;
          warray[warr_curr].fptr = &originx;
          setupstring(warr_curr++);

      boxbly = XtVaCreateManagedWidget("boxbly",
                                       formWidgetClass,
                                       boxbl,
                                       XtNfromHoriz, boxblx,
                                       XtNfromVert, labbl,
                                       XtNborderWidth, 0,
                                       NULL);
          labbly = XtVaCreateManagedWidget("y=",
                                           labelWidgetClass,
                                           boxbly,
                                           XtNborderWidth, 0,
                                           NULL);
          txtbly = XtVaCreateManagedWidget("txtbly",
                                           asciiTextWidgetClass,
                                           boxbly,
                                           XtNeditType, XawtextEdit,
                                           XtNresizable, True,
                                           XtNresize, XawtextResizeBoth,
                                           XtNfromHoriz, labbly,
                                           XtNright, XtChainRight,
                                           XtNleft, XtChainLeft,
                                           XtNwidth, 80,
                                           XtNdisplayCaret, False,
                                           NULL);
          warray[warr_curr].w = txtbly;
          warray[warr_curr].type = FLOAT_TYPE;
          warray[warr_curr].fptr = &originy;
          setupstring(warr_curr++);

  boxtr = XtVaCreateManagedWidget("boxtr",
                                  formWidgetClass,
                                  optbox,
                                  NULL);
      labtr = XtVaCreateManagedWidget("Top right of graph",
                                      labelWidgetClass,
                                      boxtr,
                                      XtNborderWidth, 0,
                                      NULL);
      boxtrx = XtVaCreateManagedWidget("boxtrx",
                                       formWidgetClass,
                                       boxtr,
                                       XtNfromVert, labtr,
                                       XtNborderWidth, 0,
                                       NULL);
          labtrx = XtVaCreateManagedWidget("x=",
                                           labelWidgetClass,
                                           boxtrx,
                                           XtNborderWidth, 0,
                                           NULL);
          txttrx = XtVaCreateManagedWidget("txttrx",
                                           asciiTextWidgetClass,
                                           boxtrx,
                                           XtNeditType, XawtextEdit,
                                           XtNresizable, True,
                                           XtNresize, XawtextResizeBoth,
                                           XtNfromHoriz, labtrx,
                                           XtNright, XtChainRight,
                                           XtNleft, XtChainLeft,
                                           XtNwidth, 80,
                                           XtNdisplayCaret, False,
                                           NULL);
          warray[warr_curr].w = txttrx;
          warray[warr_curr].type = FLOAT_TYPE;
          warray[warr_curr].fptr = &pagewidth;
          setupstring(warr_curr++);

      boxtry = XtVaCreateManagedWidget("boxtry",
                                       formWidgetClass,
                                       boxtr,
                                       XtNfromHoriz, boxtrx,
                                       XtNfromVert, labtr,
                                       XtNborderWidth, 0,
                                       NULL);
          labtry = XtVaCreateManagedWidget("y=",
                                           labelWidgetClass,
                                           boxtry,
                                           XtNborderWidth, 0,
                                           NULL);
          txttry = XtVaCreateManagedWidget("txttry",
                                           asciiTextWidgetClass,
                                           boxtry,
                                           XtNeditType, XawtextEdit,
                                           XtNresizable, True,
                                           XtNresize, XawtextResizeBoth,
                                           XtNfromHoriz, labtry,
                                           XtNright, XtChainRight,
                                           XtNleft, XtChainLeft,
                                           XtNwidth, 80,
                                           XtNdisplayCaret, False,
                                           NULL);
          warray[warr_curr].w = txttry;
          warray[warr_curr].type = FLOAT_TYPE;
          warray[warr_curr].fptr = &voffmax;
          setupstring(warr_curr++);

  boxstep = XtVaCreateManagedWidget("boxstep",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labstep = XtVaCreateManagedWidget("Step interval",
                                        labelWidgetClass,
                                        boxstep,
                                        XtNborderWidth, 0,
                                        NULL);
      txtstep = XtVaCreateManagedWidget("txtstep",
                                        asciiTextWidgetClass,
                                        boxstep,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labstep,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtstep;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &vstep;
      setupstring(warr_curr++);

  boxnorm = XtVaCreateManagedWidget("boxnorm",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labnorm = XtVaCreateManagedWidget("Norm scaling factor",
                                        labelWidgetClass,
                                        boxnorm,
                                        XtNborderWidth, 0,
                                        NULL);
      txtnorm = XtVaCreateManagedWidget("txtnorm",
                                        asciiTextWidgetClass,
                                        boxnorm,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labnorm,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtnorm;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &uscaling;
      setupstring(warr_curr++);

  boxnscl = XtVaCreateManagedWidget("boxnscl",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labnscl = XtVaCreateManagedWidget("Size of norm scale",
                                        labelWidgetClass,
                                        boxnscl,
                                        XtNborderWidth, 0,
                                        NULL);
      txtnscl = XtVaCreateManagedWidget("txtnscl",
                                        asciiTextWidgetClass,
                                        boxnscl,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labnscl,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtnscl;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &maxnorm;
      setupstring(warr_curr++);

  boxfsiz = XtVaCreateManagedWidget("boxfsiz",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labfsiz = XtVaCreateManagedWidget("Font size",
                                        labelWidgetClass,
                                        boxfsiz,
                                        XtNborderWidth, 0,
                                        NULL);
      txtfsiz = XtVaCreateManagedWidget("txtfsiz",
                                        asciiTextWidgetClass,
                                        boxfsiz,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labfsiz,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtfsiz;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &fontsize;
      setupstring(warr_curr++);

  boxtick = XtVaCreateManagedWidget("boxtick",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labtick = XtVaCreateManagedWidget("Tick interval",
                                        labelWidgetClass,
                                        boxtick,
                                        XtNborderWidth, 0,
                                        NULL);
      txttick = XtVaCreateManagedWidget("txttick",
                                        asciiTextWidgetClass, boxtick,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labtick,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txttick;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &printtime;
      setupstring(warr_curr++);

  boxmtin = XtVaCreateManagedWidget("boxmtin",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labmtin = XtVaCreateManagedWidget("Minimum t interval",
                                        labelWidgetClass,
                                        boxmtin,
                                        XtNborderWidth, 0,
                                        NULL);
      txtmtin = XtVaCreateManagedWidget("txtmtin",
                                        asciiTextWidgetClass,
                                        boxmtin,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labmtin,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtmtin;
      warray[warr_curr].type = FLOAT_TYPE;
      warray[warr_curr].fptr = &mintime;
      setupstring(warr_curr++);

  boxncol = XtVaCreateManagedWidget("boxncol",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labncol = XtVaCreateManagedWidget("No columns/page",
                                        labelWidgetClass,
                                        boxncol,
                                        XtNborderWidth, 0,
                                        NULL);
      txtncol = XtVaCreateManagedWidget("txtncol",
                                        asciiTextWidgetClass,
                                        boxncol,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labncol,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtncol;
      warray[warr_curr].type = INT_TYPE;
      warray[warr_curr].iptr = &idoubled;
      setupstring(warr_curr++);

  boxmaxp = XtVaCreateManagedWidget("boxmaxp",
                                    formWidgetClass,
                                    optbox,
                                    XtNborderWidth, 0,
                                    NULL);
      labmaxp = XtVaCreateManagedWidget("Max. no. pages",
                                        labelWidgetClass,
                                        boxmaxp,
                                        XtNborderWidth, 0,
                                        NULL);
      txtmaxp = XtVaCreateManagedWidget("txtmaxp",
                                        asciiTextWidgetClass,
                                        boxmaxp,
                                        XtNeditType, XawtextEdit,
                                        XtNresizable, True,
                                        XtNresize, XawtextResizeBoth,
                                        XtNfromHoriz, labmaxp,
                                        XtNright, XtChainRight,
                                        XtNleft, XtChainLeft,
                                        XtNdisplayCaret, False,
                                        NULL);
      warray[warr_curr].w = txtmaxp;
      warray[warr_curr].type = INT_TYPE;
      warray[warr_curr].iptr = &nopages;
      setupstring(warr_curr++);


  /* Now set up the translatations for all of the asciiTextWidgets */
  for (ii=0; ii<warrl_curr; ii++) {
    if (!warrayl[ii].notrans) {
      updatetranslations = XtParseTranslationTable(updatetranstabl);
      XtOverrideTranslations(warrayl[ii].w, updatetranslations);
    }
  }
  for (ii=0; ii<warr_curr; ii++) {
      updatetranslations = XtParseTranslationTable(updatetranstab);
      XtOverrideTranslations(warray[ii].w, updatetranslations);
  }

  boxoutr = XtVaCreateManagedWidget("boxoutr",
                                    formWidgetClass,
                                    optbox,
                                    NULL);
      togfunc = XtVaCreateManagedWidget("    Functional output   ",
                                        toggleWidgetClass,
                                        boxoutr,
                                        XtNradioData, &(val[0]),
                                        XtNstate, (grayscale==0),
                                        NULL);
      XtAddCallback(togfunc, XtNcallback, Outfunc, 0);
      toggray = XtVaCreateManagedWidget("     Grayscale output   ",
                                        toggleWidgetClass,
                                        boxoutr,
                                        XtNradioData, &(val[1]),
                                        XtNradioGroup, togfunc,
                                        XtNfromVert, togfunc,
                                        XtNstate, (grayscale==1),
                                        NULL);
      XtAddCallback(toggray, XtNcallback, Outfunc, (XtPointer) 1);
      togrgry = XtVaCreateManagedWidget("Inverse grayscale output",
                                        toggleWidgetClass,
                                        boxoutr,
                                        XtNradioData, &(val[2]),
                                        XtNradioGroup, togfunc,
                                        XtNfromVert, toggray,
                                        XtNstate, (grayscale==-1),
                                        NULL);
      XtAddCallback(togrgry, XtNcallback, Outfunc, (XtPointer) 2);
      /* change the translation table so that this a precisely -one active */
      /* radio group */
      radiotrans = XtParseTranslationTable(radiotranstab);
      XtOverrideTranslations(togfunc, radiotrans);
      XtOverrideTranslations(togrgry, radiotrans);
      XtOverrideTranslations(toggray, radiotrans);

  togtimr = XtVaCreateManagedWidget("Time reverse output",
                                    toggleWidgetClass,
                                    optbox,
                                    XtNstate, (reverse==-1),
                                    NULL);
  XtAddCallback(togtimr, XtNcallback, Reverse, 0);

  redraw = XtVaCreateManagedWidget("Redraw",
                                   commandWidgetClass,
                                   optbox,
                                   NULL);
  XtAddCallback(redraw, XtNcallback, Redraw, 0);

  quit = XtVaCreateManagedWidget("Quit",
                                 commandWidgetClass,
                                 optbox,
                                 NULL);
  XtAddCallback(quit, XtNcallback, Quit, 0);

  /**************************************************************************/
  /* Realize the widgets                                                    */
  /**************************************************************************/
  XtRealizeWidget(topLevel);

  XtAppMainLoop(app_context);
#endif
}
